﻿using System.Net;
using System.Net.Http;
using System.Threading.Tasks;
using FluentAssertions;
using Microsoft.Extensions.Caching.Memory;
using NSubstitute;
using Xunit;

namespace Robware.Lib.ApiClient.Tests {
	public class ApiClientTests {
		private class TestApiClient : ApiClient {
			public TestApiClient(HttpClient client, IMemoryCache cache) : base(client, cache, new CacheDurations()) {
			}

			public async Task<T> Post<T>(string url, object value, object query = null) => await base.Post<T>(url, value, query);

			public async Task<T> Get<T>(string url, object query = null, bool skipCache = false) => await base.Get<T>(url, query, skipCache);
		}

		private class TestObject {
			public int TestProperty { get; set; }
		}

		[Fact]
		public async Task Get_WithUrl_CachesAndReturnsResult() {
			var httpClient = new HttpClientBuilder()
			                 .WithMethod(HttpMethod.Get)
			                 .WithUrl("/test")
			                 .WithResponse(@"{""TestProperty"":1}")
			                 .Build();

			var cache = Substitute.For<IMemoryCache>();
			cache.TryGetValue("TestApiClient:Get/test", out _).Returns(false);

			var expected = new TestObject {TestProperty = 1};

			var client = new TestApiClient(httpClient, cache);
			(await client.Get<TestObject>("test")).Should().BeEquivalentTo(expected);
			cache.Received(1).CreateEntry("TestApiClient:Get/test");
		}

		[Fact]
		public async Task Get_WithUrlAndQuery_CachesAndReturnsResult() {
			var httpClient = new HttpClientBuilder()
			                 .WithMethod(HttpMethod.Get)
			                 .WithUrl("/test")
			                 .WithQueryString("query1", "1")
			                 .WithQueryString("query2", "2")
			                 .WithResponse(@"{""TestProperty"":1}")
			                 .Build();

			var cache = Substitute.For<IMemoryCache>();
			cache.TryGetValue("TestApiClient:Get/test?query1=1&query2=2", out _).Returns(false);

			var expected = new TestObject {TestProperty = 1};

			var client = new TestApiClient(httpClient, cache);
			(await client.Get<TestObject>("test", new {query1 = 1, query2 = 2})).Should().BeEquivalentTo(expected);
			cache.Received(1).CreateEntry("TestApiClient:Get/test?query1=1&query2=2");
		}

		[Fact]
		public async void Get_WithUrlThatReturnsError_ThrowsException() {
			var httpClient = new HttpClientBuilder()
			                 .WithMethod(HttpMethod.Get)
			                 .WithUrl("/test")
			                 .WithResponse("")
			                 .WithErrorStatus(HttpStatusCode.NotFound)
			                 .Build();

			var cache = Substitute.For<IMemoryCache>();

			var client = new TestApiClient(httpClient, cache);

			await client.Invoking(apiClient => apiClient.Get<TestObject>("404")).Should().ThrowAsync<ApiCallException>()
			      .WithMessage("Error calling API http://example.com/404: 404, Not Found");
		}

		[Fact]
		public async Task Get_WithUrlAndCacheSkipFlag_SkipsCacheAndReturnsResult() {
			var httpClient = new HttpClientBuilder()
							 .WithMethod(HttpMethod.Get)
							 .WithUrl("/test")
							 .WithResponse(@"{""TestProperty"":1}")
							 .Build();

			var cache = Substitute.For<IMemoryCache>();
			cache.TryGetValue("TestApiClient:Get/test", out _).Returns(false);

			var expected = new TestObject { TestProperty = 1 };

			var client = new TestApiClient(httpClient, cache);
			(await client.Get<TestObject>("test", skipCache: true)).Should().BeEquivalentTo(expected);
			cache.Received(0).CreateEntry("TestApiClient:Get/test");
		}

		[Fact]
		public async Task Post_WithUrlAndValue_DoesNotCacheAndReturnsResult() {
			var httpClient = new HttpClientBuilder()
			                 .WithMethod(HttpMethod.Post)
			                 .WithUrl("/test")
			                 .WithResponse(@"{""TestProperty"":1}")
			                 .WithPostBody("\"value\"")
			                 .Build();

			var cache = Substitute.For<IMemoryCache>();

			var expected = new TestObject {TestProperty = 1};

			var client = new TestApiClient(httpClient, cache);
			(await client.Post<TestObject>("test", "value")).Should().BeEquivalentTo(expected);
			cache.Received(0).CreateEntry(Arg.Any<string>());
		}

		[Fact]
		public async Task Post_WithUrlAndValueAndQuery_DoesNotCacheAndReturnsResult() {
			var httpClient = new HttpClientBuilder()
			                 .WithMethod(HttpMethod.Post)
			                 .WithUrl("/test")
			                 .WithQueryString("query1", "1")
			                 .WithQueryString("query2", "2")
			                 .WithPostBody("\"value\"")
							 .WithResponse(@"{""TestProperty"":1}")
			                 .Build();

			var cache = Substitute.For<IMemoryCache>();

			var expected = new TestObject {TestProperty = 1};

			var client = new TestApiClient(httpClient, cache);
			(await client.Post<TestObject>("test", "value", new {query1 = 1, query2 = 2})).Should().BeEquivalentTo(expected);
			cache.Received(0).CreateEntry(Arg.Any<string>());
		}

		[Fact]
		public async Task Post_WithUrlAndValueAndQuery_DoesNotCacheAndReturnsNoResult() {
			var httpClient = new HttpClientBuilder()
			                 .WithMethod(HttpMethod.Post)
			                 .WithUrl("/test")
			                 .WithQueryString("query1", "1")
			                 .WithQueryString("query2", "2")
			                 .WithPostBody("\"value\"")
			                 .Build(out var mockHttpMessageHandler);

			var cache = Substitute.For<IMemoryCache>();

			var client = new TestApiClient(httpClient, cache);
			await client.Post<object>("test", "value", new {query1 = 1, query2 = 2});
			mockHttpMessageHandler.VerifyNoOutstandingExpectation();
			cache.Received(0).CreateEntry(Arg.Any<string>());
		}
	}
}
