﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.Http;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;
using Microsoft.AspNetCore.WebUtilities;
using Microsoft.Extensions.Caching.Memory;
using Newtonsoft.Json;

namespace Robware.Lib.ApiClient {
	public abstract class ApiClient {
		private readonly HttpClient _client;
		private readonly IMemoryCache _cache;
		private readonly CacheDurations _cacheDurations;

		protected ApiClient(HttpClient client, IMemoryCache cache, CacheDurations cacheDurations) {
			_client = client;
			_cache = cache;
			_cacheDurations = cacheDurations;
		}

		private static IDictionary<string, string> ParseQueryParameters(object query) {
			var type = query.GetType();
			var props = type.GetProperties();
			return props.ToDictionary(info => info.Name, info => info.GetValue(query, null)?.ToString());
		}

		private async Task<TResult> DoCachedCall<TResult>(string callerName, string url, bool skipCache, Func<Task<TResult>> call) {
			if (skipCache)
				return await call();

			var baseKey = GetType().Name + ":" + callerName;
			var key = baseKey + "/" + url;

			return await _cache.GetOrCreate(key, async entry => {
				entry.AbsoluteExpirationRelativeToNow = TimeSpan.FromSeconds(_cacheDurations[baseKey]);
				return await call();
			});
		}

		private async Task<T> Send<T>(HttpMethod method, string url, object query, bool skipCache, string callerName, HttpContent content = null) {
			if (query != null)
				url = QueryHelpers.AddQueryString(url, ParseQueryParameters(query));

			return await DoCachedCall(callerName, url, skipCache, async () => {
				using var httpRequest = new HttpRequestMessage(method, url) { Content = content };
				var response = await _client.SendAsync(httpRequest);

				if (!response.IsSuccessStatusCode)
					throw new ApiCallException(response);

				var json = await response.Content.ReadAsStringAsync();
				return JsonConvert.DeserializeObject<T>(json);
			});
		}

		private async Task<T> Upload<T>(HttpMethod method, string url, object value, object query, string callerName) {
			var json = JsonConvert.SerializeObject(value);
			using var requestBody = new StringContent(json, Encoding.UTF8, "application/json");
			return await Send<T>(method, url, query, true, callerName, requestBody);
		}
		
		protected async Task<T> Post<T>(string url, object value, object query = null, [CallerMemberName] string callerName = "") =>
			await Upload<T>(HttpMethod.Post, url, value, query, callerName);

		protected async Task<T> Put<T>(string url, object value, object query = null, [CallerMemberName] string callerName = "") =>
			await Upload<T>(HttpMethod.Put, url, value, query, callerName);

		protected async Task<T> Patch<T>(string url, object value, object query = null, [CallerMemberName] string callerName = "") => 
			await Upload<T>(HttpMethod.Patch, url, value, query, callerName);

		protected async Task<T> Get<T>(string url, object query = null, bool skipCache = false, [CallerMemberName] string callerName = "") => 
			await Send<T>(HttpMethod.Get, url, query, skipCache, callerName);

		protected async Task<T> Delete<T>(string url, object query = null, bool skipCache = false, [CallerMemberName] string callerName = "") => 
			await Send<T>(HttpMethod.Delete, url, query, skipCache, callerName);
	}
}
